/*
Copyright 2020 The Kruise Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package v1alpha1

import (
	"fmt"

	"github.com/openkruise/kruise/apis/apps/v1beta1"
	"sigs.k8s.io/controller-runtime/pkg/conversion"
)

func (sts *StatefulSet) ConvertTo(dst conversion.Hub) error {
	switch t := dst.(type) {
	case *v1beta1.StatefulSet:
		stsv1beta1 := dst.(*v1beta1.StatefulSet)
		stsv1beta1.ObjectMeta = sts.ObjectMeta

		// spec
		stsv1beta1.Spec = v1beta1.StatefulSetSpec{
			Replicas:             sts.Spec.Replicas,
			Selector:             sts.Spec.Selector,
			Template:             sts.Spec.Template,
			VolumeClaimTemplates: sts.Spec.VolumeClaimTemplates,
			ServiceName:          sts.Spec.ServiceName,
			PodManagementPolicy:  sts.Spec.PodManagementPolicy,
			UpdateStrategy: v1beta1.StatefulSetUpdateStrategy{
				Type: sts.Spec.UpdateStrategy.Type,
			},
			RevisionHistoryLimit: sts.Spec.RevisionHistoryLimit,
		}
		if sts.Spec.UpdateStrategy.RollingUpdate != nil {
			stsv1beta1.Spec.UpdateStrategy.RollingUpdate = &v1beta1.RollingUpdateStatefulSetStrategy{
				Partition:             sts.Spec.UpdateStrategy.RollingUpdate.Partition,
				MaxUnavailable:        sts.Spec.UpdateStrategy.RollingUpdate.MaxUnavailable,
				PodUpdatePolicy:       v1beta1.PodUpdateStrategyType(sts.Spec.UpdateStrategy.RollingUpdate.PodUpdatePolicy),
				Paused:                sts.Spec.UpdateStrategy.RollingUpdate.Paused,
				InPlaceUpdateStrategy: sts.Spec.UpdateStrategy.RollingUpdate.InPlaceUpdateStrategy,
				MinReadySeconds:       sts.Spec.UpdateStrategy.RollingUpdate.MinReadySeconds,
			}
			if sts.Spec.UpdateStrategy.RollingUpdate.UnorderedUpdate != nil {
				stsv1beta1.Spec.UpdateStrategy.RollingUpdate.UnorderedUpdate = &v1beta1.UnorderedUpdateStrategy{
					PriorityStrategy: sts.Spec.UpdateStrategy.RollingUpdate.UnorderedUpdate.PriorityStrategy,
				}
			}
		}

		// status
		stsv1beta1.Status = v1beta1.StatefulSetStatus{
			ObservedGeneration: sts.Status.ObservedGeneration,
			Replicas:           sts.Status.Replicas,
			ReadyReplicas:      sts.Status.ReadyReplicas,
			AvailableReplicas:  sts.Status.AvailableReplicas,
			CurrentReplicas:    sts.Status.CurrentReplicas,
			UpdatedReplicas:    sts.Status.UpdatedReplicas,
			CurrentRevision:    sts.Status.CurrentRevision,
			UpdateRevision:     sts.Status.UpdateRevision,
			CollisionCount:     sts.Status.CollisionCount,
			Conditions:         sts.Status.Conditions,
			LabelSelector:      sts.Status.LabelSelector,
		}

		return nil

	default:
		return fmt.Errorf("unsupported type %v", t)
	}
}

func (sts *StatefulSet) ConvertFrom(src conversion.Hub) error {
	switch t := src.(type) {
	case *v1beta1.StatefulSet:
		stsv1beta1 := src.(*v1beta1.StatefulSet)
		sts.ObjectMeta = stsv1beta1.ObjectMeta

		// spec
		sts.Spec = StatefulSetSpec{
			Replicas:             stsv1beta1.Spec.Replicas,
			Selector:             stsv1beta1.Spec.Selector,
			Template:             stsv1beta1.Spec.Template,
			VolumeClaimTemplates: stsv1beta1.Spec.VolumeClaimTemplates,
			ServiceName:          stsv1beta1.Spec.ServiceName,
			PodManagementPolicy:  stsv1beta1.Spec.PodManagementPolicy,
			UpdateStrategy: StatefulSetUpdateStrategy{
				Type: stsv1beta1.Spec.UpdateStrategy.Type,
			},
			RevisionHistoryLimit: stsv1beta1.Spec.RevisionHistoryLimit,
		}
		if stsv1beta1.Spec.UpdateStrategy.RollingUpdate != nil {
			sts.Spec.UpdateStrategy.RollingUpdate = &RollingUpdateStatefulSetStrategy{
				Partition:             stsv1beta1.Spec.UpdateStrategy.RollingUpdate.Partition,
				MaxUnavailable:        stsv1beta1.Spec.UpdateStrategy.RollingUpdate.MaxUnavailable,
				PodUpdatePolicy:       PodUpdateStrategyType(stsv1beta1.Spec.UpdateStrategy.RollingUpdate.PodUpdatePolicy),
				Paused:                stsv1beta1.Spec.UpdateStrategy.RollingUpdate.Paused,
				InPlaceUpdateStrategy: stsv1beta1.Spec.UpdateStrategy.RollingUpdate.InPlaceUpdateStrategy,
				MinReadySeconds:       stsv1beta1.Spec.UpdateStrategy.RollingUpdate.MinReadySeconds,
			}
			if stsv1beta1.Spec.UpdateStrategy.RollingUpdate.UnorderedUpdate != nil {
				sts.Spec.UpdateStrategy.RollingUpdate.UnorderedUpdate = &UnorderedUpdateStrategy{
					PriorityStrategy: stsv1beta1.Spec.UpdateStrategy.RollingUpdate.UnorderedUpdate.PriorityStrategy,
				}
			}
		}

		// status
		sts.Status = StatefulSetStatus{
			ObservedGeneration: stsv1beta1.Status.ObservedGeneration,
			Replicas:           stsv1beta1.Status.Replicas,
			ReadyReplicas:      stsv1beta1.Status.ReadyReplicas,
			AvailableReplicas:  stsv1beta1.Status.AvailableReplicas,
			CurrentReplicas:    stsv1beta1.Status.CurrentReplicas,
			UpdatedReplicas:    stsv1beta1.Status.UpdatedReplicas,
			CurrentRevision:    stsv1beta1.Status.CurrentRevision,
			UpdateRevision:     stsv1beta1.Status.UpdateRevision,
			CollisionCount:     stsv1beta1.Status.CollisionCount,
			Conditions:         stsv1beta1.Status.Conditions,
			LabelSelector:      stsv1beta1.Status.LabelSelector,
		}

		return nil
	default:
		return fmt.Errorf("unsupported type %v", t)
	}
}
